#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import scipy
from ._op_reqs import *

"""
Reduction Op Superclasses
"""


class ReductionAxes(Operation):
    input_spec = InputSpec(
        x=TensorInputType(),
        axes=IntTensorInputType(const=True, optional=True),
        keep_dims=BoolInputType(const=True, optional=True),
    )

    def default_inputs(self):
        return DefaultInputs(
            axes=None,
            keep_dims=False,
            )

    def __init__(self, **kwargs):
        super(ReductionAxes, self).__init__(**kwargs)

    def type_inference(self):
        x_type = self.x.dtype
        x_shape = self.x.shape
        axes = self.axes.val if self.axes is not None else None
        if axes is None:
            axes = range(self.x.rank)
        keep_dims = self.keep_dims.val

        reduced_shape = list(x_shape)
        if keep_dims:
            for i in axes:
                reduced_shape[i] = 1
        else:
            # sort reverse so we can delete shape elements back to front
            axes = [axis if axis >= 0 else axis + len(reduced_shape) for axis in axes]
            for i in sorted(axes)[::-1]:
                reduced_shape.pop(i)
        if len(reduced_shape) == 0:
            return x_type  # scalar

        return types.tensor(x_type, tuple(reduced_shape))

    @precondition(allow=VALUE)
    def value_inference(self):
        axes = tuple(self.axes.val) if self.axes is not None else None
        return self.get_operator()(self.x.val, axis=axes, keepdims=self.keep_dims.val)

    def get_operator(self):
        raise NotImplementedError()


class ReductionAxis(Operation):
    input_spec = InputSpec(
        x=TensorInputType(),
        axis=IntInputType(const=True, optional=True),
        keep_dims=BoolInputType(const=True, optional=True),
    )

    def default_inputs(self):
        return DefaultInputs(
            axis=-1,
            keep_dims=False,
            )

    def __init__(self, **kwargs):
        super(ReductionAxis, self).__init__(**kwargs)

    def _find_reduced_shape(self):
        x_shape = self.x.shape
        axis = self.axis.val

        reduced_shape = list(x_shape)
        axis = axis if axis >= 0 else axis + len(reduced_shape)
        if self.keep_dims.val:
            reduced_shape[axis] = 1
        else:
            reduced_shape.pop(axis)
        return reduced_shape

    def type_inference(self):
        x_type = self.x.dtype
        reduced_shape = self._find_reduced_shape_and_axis()
        return types.tensor(x_type, tuple(reduced_shape))

    @precondition(allow=VALUE)
    def value_inference(self):
        tmp = self.get_operator()(self.x.val, axis=self.axis.val)
        reduced_shape = self._find_reduced_shape()
        if self.keep_dims.val:
            tmp = np.reshape(tmp, reduced_shape)
        return tmp

    def get_operator(self):
        raise NotImplementedError()


"""
Reduction op implementations
"""


@register_op(doc_str="TODO")
class reduce_arg(ReductionAxis):
    def __init__(self, **kwargs):
        super(reduce_arg, self).__init__(**kwargs)

    def type_inference(self):
        x_shape = self.x.shape
        axis = self.axis.val

        reduced_shape = list(x_shape)
        axis = axis if axis >= 0 else axis + len(reduced_shape)
        if self.keep_dims.val:
            reduced_shape[axis] = 1
        else:
            reduced_shape.pop(axis)

        return types.tensor(types.int32, tuple(reduced_shape))


@register_op(doc_str="")
class reduce_argmax(reduce_arg):
    """
    Computes the indices of the maximum value across dimensions of a tensor.
    In case of ties, the identity of the return value is not guaranteed.
    
    Parameters
    ----------
    x: <\*,T> (Required)
        * Must be 1-dimensional or higher.
    
    axis: const<i32> (Optional)
        * The dimension to reduce. Default is ``-1``.
    
    keep_dims: const<bool> (Optional, default=False)
        * If ``False``, the rank is reduced by ``1`` by removing the dimension
          specified in ``axis``. If ``True``, retain reduced axis with length ``1``.
    
    Returns
    -------
    <\*, int32>
    
    Attributes
    ----------
    T: f32, int32
    
    References
    ----------
    See `tf.math.argmax <https://www.tensorflow.org/api_docs/python/tf/math/argmax>`_.
    
    """
    
    def __init__(self, **kwargs):
        super(reduce_argmax, self).__init__(**kwargs)

    def get_operator(self):
        return np.argmax


@register_op(doc_str="")
class reduce_argmin(reduce_arg):
    """
    Computes the indices of the minimum value across dimensions of a tensor.
    In case of ties, the identity of the return value is not guaranteed.
    
    Parameters
    ----------
    x: <\*,T> (Required)
        * Must be 1-dimensional or higher.
    
    axis: const<i32> (Optional)
        * The dimension to reduce. Default is ``-1``.
    
    keep_dims: const<bool> (Optional, default=False)
        * If ``False``, the rank is reduced by ``1`` by removing the dimension specified
          in ``axis``, otherwise retain reduced axis with length ``1``.
    
    Returns
    -------
    <\*, int32>
    
    Attributes
    ----------
    T: f32, int32
    
    References
    ----------
    See `tf.math.argmin <https://www.tensorflow.org/api_docs/python/tf/math/argmin>`_.
    
    """
    
    def __init__(self, **kwargs):
        super(reduce_argmin, self).__init__(**kwargs)

    def get_operator(self):
        return np.argmin


@register_op(doc_str="")
class reduce_l1_norm(ReductionAxes):
    """
    Computes the L1 normalization of elements across given dimensions of the input tensor.
    
    Parameters
    ----------
    x: <\*,T> (Required)
        * Must be 1-dimensional or higher.
    
    axes: const<K,i32> (Optional, default="None", reduce on all axes.)
        * The dimensions to reduce.
    
    keep_dims: const<bool> (Optional, default=False)
        * If ``False``, the rank is reduced by ``1`` for each entry in ``axes``,
          otherwise retain reduced axes with length ``1``.
    
    Returns
    -------
    <\*,T>
        * Scalar or tensor: The reduced tensor.
    
    Attributes
    ----------
    T: f32, int32
    
    References
    ----------
    See `reduce_mean <https://www.tensorflow.org/api_docs/python/tf/math/reduce_mean?version=stable>`_.
    
    """
    
    def __init__(self, **kwargs):
        super(reduce_l1_norm, self).__init__(**kwargs)

    def get_operator(self):
        def l1_norm(x, axis=None, keepdims=False):
            return np.sum(np.abs(x), axis=axis, keepdims=keepdims)

        return l1_norm


@register_op(doc_str="")
class reduce_l2_norm(ReductionAxes):
    """
    Computes the L2 normalization of elements across given dimensions of the input tensor.
    
    Parameters
    ----------
    x: <\*,T> (Required)
        * Must be 1-dimensional or higher.
    
    axes: const<K,i32> (Optional, default="None", reduce on all axes.)
        * The dimensions to reduce.
    
    keep_dims: const<bool> (Optional, default=False)
        * If ``False``, the rank is reduced by ``1`` for each entry in ``axes``,
          otherwise retain reduced axes with length ``1``.
    
    Returns
    -------
    <\*,T>
        * Scalar or tensor: The reduced tensor.
    
    Attributes
    ----------
    T: f32, int32
    
    """
    
    def __init__(self, **kwargs):
        super(reduce_l2_norm, self).__init__(**kwargs)

    def get_operator(self):
        def l2_norm(x, axis=None, keepdims=False):
            return np.sqrt(np.sum(np.square(x), axis=axis, keepdims=keepdims))

        return l2_norm


@register_op(doc_str="")
class reduce_log_sum(ReductionAxes):
    """
    Computes the natural logarithm of the sum of all the elements across given dimensions
    of the input tensor.
    
    Parameters
    ----------
    x: <\*,T> (Required)
        * Must be 1-dimensional or higher.
    
    axes: const<K,i32> (Optional, default="None", reduce on all axes.)
        * The dimensions to reduce.
    
    keep_dims: const<bool> (Optional, default=False)
        * If ``False``, the rank is reduced by ``1`` for each entry in ``axes``,
          otherwise retain reduced axes with length ``1``.
    
    Returns
    -------
    <\*,T>
        * Scalar or tensor: The reduced tensor.
    
    Attributes
    ----------
    T: f32, int32
    
    """
    
    def __init__(self, **kwargs):
        super(reduce_log_sum, self).__init__(**kwargs)

    def get_operator(self):
        def log_sum(x, axis=None, keepdims=False):
            return np.log(np.sum(x, axis=axis, keepdims=keepdims))

        return log_sum


@register_op(doc_str="")
class reduce_log_sum_exp(ReductionAxes):
    """
    Computes the natural logarithm of the sum of the exponentials of the elements across
    given dimensions of the input tensor. It is a smooth approximation of the maximum
    function, more numerically stable than ``log(sum(exp(input)))``. It avoids
    overflows caused by taking the ``exp`` of large inputs and underflows caused by
    taking the ``log`` of small inputs.
    
    Parameters
    ----------
    x: <\*,T> (Required)
        * Must be 1-dimensional or higher.
    
    axes: const<K,i32> (Optional, default="None", reduce on all axes.)
        * The dimensions to reduce.
    
    keep_dims: const<bool> (Optional, default=False)
        * If ``False``, the rank is reduced by ``1`` for each entry in ``axes``,
          otherwise retain reduced axes with length ``1``.
    
    Returns
    -------
    <\*,T>
        * Scalar or tensor: The reduced tensor.
    
    Attributes
    ----------
    T: f32, int32
    
    References
    ----------
    See `tf.math.reduce_logsumexp <https://www.tensorflow.org/api_docs/python/tf/math/reduce_logsumexp>`_.
    
    """
    
    def __init__(self, **kwargs):
        super(reduce_log_sum_exp, self).__init__(**kwargs)

    def get_operator(self):
        return scipy.special.logsumexp


@register_op(doc_str="")
class reduce_max(ReductionAxes):
    """
    Computes the maximum of elements across given dimensions of the input tensor.
    
    Parameters
    ----------
    x: <\*,T> (Required)
        * Must be 1-dimensional or higher.
    
    axes: const<K,i32> (Optional, default="None", reduce on all axes.)
        * The dimensions to reduce.
    
    keep_dims: const<bool> (Optional, default=False)
        * If ``False``, the rank is reduced by ``1`` for each entry in ``axes``,
          otherwise retain reduced axes with length ``1``.
    
    Returns
    -------
    <\*,T>
        * Scalar or tensor: The reduced tensor.
    
    Attributes
    ----------
    T: f32, int32
    
    """
    
    def __init__(self, **kwargs):
        super(reduce_max, self).__init__(**kwargs)

    def get_operator(self):
        return np.max


@register_op(doc_str="")
class reduce_mean(ReductionAxes):
    """
    Computes the mean of elements across given dimensions of the input tensor.
    
    Parameters
    ----------
    x: <\*,T> (Required)
        * Must be 1-dimensional or higher.
    
    axes: const<K,i32> (Optional, default="None", reduce on all axes.)
        * The dimensions to reduce.
    
    keep_dims: const<bool> (Optional, default=False)
        * If ``False``, the rank is reduced by ``1`` for each entry in ``axes``,
          otherwise retain reduced axes with length ``1``.
    
    Returns
    -------
    <\*,T>
        * Scalar or tensor: The reduced tensor.
    
    Attributes
    ----------
    T: f32, int32
    
    References
    ----------
    For an example, see `tf.math.reduce_mean <https://www.tensorflow.org/api_docs/python/tf/math/reduce_mean?version=stable>`_.
    """
    
    def __init__(self, **kwargs):
        super(reduce_mean, self).__init__(**kwargs)

    def get_operator(self):
        return np.mean


@register_op(doc_str="")
class reduce_min(ReductionAxes):
    """
    Computes the minimum of elements across given dimensions of the input tensor.
    
    Parameters
    ----------
    x: <\*,T> (Required)
        * Must be 1-dimensional or higher.
    
    axes: const<K,i32> (Optional, default="None", reduce on all axes.)
        * The dimensions to reduce.
    
    keep_dims: const<bool> (Optional, default=False)
        * If ``False``, the rank is reduced by ``1`` for each entry in ``axes``,
          otherwise retain reduced axes with length ``1``.
    
    Returns
    -------
    <\*,T>
        * Scalar or tensor: The reduced tensor.
    
    Attributes
    ----------
    T: f32, int32
    
    """
    
    def __init__(self, **kwargs):
        super(reduce_min, self).__init__(**kwargs)

    def get_operator(self):
        return np.min


@register_op(doc_str="")
class reduce_prod(ReductionAxes):
    """
    Computes the product of elements across given dimensions of the input tensor.
    
    Parameters
    ----------
    x: <\*,T> (Required)
        * Must be 1-dimensional or higher.
    
    axes: const<K,i32> (Optional, default="None", reduce on all axes.)
        * The dimensions to reduce.
    
    keep_dims: const<bool> (Optional, default=False)
        * If ``False``, the rank is reduced by ``1`` for each entry in ``axes``,
          otherwise retain reduced axes with length ``1``.
    
    Returns
    -------
    <\*,T>
        * Scalar or tensor: The reduced tensor.
    
    Attributes
    ----------
    T: f32, int32
    
    """
    
    def __init__(self, **kwargs):
        super(reduce_prod, self).__init__(**kwargs)

    def get_operator(self):
        return np.prod


@register_op(doc_str="")
class reduce_sum(ReductionAxes):
    """
    Computes the sum of elements across given dimensions of the input tensor.
    
    Parameters
    ----------
    x: <\*,T> (Required)
        * Must be 1-dimensional or higher.
    
    axes: const<K,i32> (Optional, default="None", reduce on all axes.)
        * The dimensions to reduce.
    
    keep_dims: const<bool> (Optional, default=False)
        * If ``False``, the rank is reduced by ``1`` for each entry in ``axes``,
          otherwise retain reduced axes with length ``1``.
    
    Returns
    -------
    <\*,T>
        * Scalar or tensor: The reduced tensor.
    
    Attributes
    ----------
    T: f32, int32
    
    """
    
    def __init__(self, **kwargs):
        super(reduce_sum, self).__init__(**kwargs)

    def get_operator(self):
        return np.sum


@register_op(doc_str="")
class reduce_sum_square(ReductionAxes):
    """
    Computes the sum of squares of elements across given dimensions of the input tensor.
    
    Parameters
    ----------
    x: <\*,T> (Required)
        * Must be 1-dimensional or higher.
    
    axes: const<K,i32> (Optional, default="None", reduce on all axes.)
        * The dimensions to reduce.
    
    keep_dims: const<bool> (Optional, default=False)
        * If ``False``, the rank is reduced by ``1`` for each entry in ``axes``,
          otherwise retain reduced axes with length ``1``.
    
    Returns
    -------
    <\*,T>
        * Scalar or tensor: The reduced tensor.
    
    Attributes
    ----------
    T: f32, int32
    
    """
    
    def __init__(self, **kwargs):
        super(reduce_sum_square, self).__init__(**kwargs)

    def get_operator(self):
        def sum_squre(x, axis=None, keepdims=False):
            return np.sum(np.square(x), axis=axis, keepdims=keepdims)

        return sum_squre



# -*- coding: utf-8 -*-

#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from __future__ import print_function as _
from __future__ import division as _
from __future__ import absolute_import as _

from coremltools.converters.mil.mil import Builder as mb
from coremltools.converters.mil.mil.passes.pass_registry import register_pass
import logging


def remove_vacuous_cond_block(block):
    num_changes = 0
    for op in list(block.operations):
        for b in op.blocks:
            num_changes += remove_vacuous_cond_block(b)

        if op.op_type != "cond":
            continue

        then_ops = op.blocks[0].operations
        else_ops = op.blocks[1].operations

        if len(then_ops) > 1 or len(else_ops) > 1:
            continue

        # Pattern 1: dynamic length TensorList generates this pattern. See
        # conversion functions of TensorList* ops for details. TF2's graph
        # contains a tf.cond op with 2 sub-graphs. The condition is either
        # `less_equal` or `greater_equal` op. 1 sub-graph contains only an
        # identity op forwarding the original TensorList, another sub-graph
        # contains TensorListResize op to generate a new TensorList. But in
        # backend, list length is handled dynamically in list_write/scatter
        # and thus, the entire tf.cond and it's sub-graphs can be removed.
        if len(then_ops) == 0 and len(else_ops) == 0:
            if op.pred.op.op_type not in {"less_equal", "greater_equal"}:
                continue

            # cond op must have pred
            pred_x = op.pred.op.x.op
            pred_y = op.pred.op.y.op

            if pred_x is None and pred_y is None:
                continue

            if op.pred.op.op_type == "less_equal":
                if pred_x.op_type != "list_length":
                    continue
                new_var = pred_x.ls

            else:  # op.pred.op.op_type == 'greather_equal':
                if pred_y.op_type != "list_length":
                    continue
                new_var = pred_y.ls

            with block:
                op.enclosing_block.replace_uses_of_var_after_op(
                    anchor_op=op, old_var=op.outputs[0], new_var=new_var
                )
                block.remove_ops([op])  # rely on DCE to remove extra cond inputs
            num_changes += 1

        # Pattern 2: both than and else branch contains exactly 1 identity op
        if len(then_ops) == 1 and len(then_ops) == 1:
            if then_ops[0].op_type != "identity" or else_ops[0].op_type != "identity":
                continue
            if then_ops[0].x != else_ops[0].x:
                continue

            with block:
                new_var = mb.identity(x=then_ops[0].x, before_op=op, name=op.name)
                op.enclosing_block.replace_uses_of_var_after_op(
                    anchor_op=op, old_var=op.outputs[0], new_var=new_var
                )
                block.remove_ops([op])  # rely on DCE to remove extra cond inputs
            num_changes += 1

    return num_changes


@register_pass(namespace="tensorflow2")
def remove_vacuous_cond(prog):
    """
    Remove cond op and it's sub-graphs that produces identity on both then and
    else branch. One example use case is the TensorListReverse op, in Core ML,
    we dynamically resize in write operations, and thus, both branches of the
    cond op will be a skip (identity) op.

    Given:

        main(%a: (1, bool),
         %b: (2, 3, fp32)) {
          block0() {
            %squeeze_0: (bool) = squeeze(x=%a, name="squeeze_0")
            %cond_0: (2, 3, fp32) = cond(pred=%squeeze_0, name="cond_0")
              cond_0_true() {
                %identity_0: (2, 3, fp32) = identity(x=%b, name="identity_0")
              } -> (%identity_0)
              cond_0_false() {
                %identity_1: (2, 3, fp32) = identity(x=%b, name="identity_1")
              } -> (%identity_1)
          } -> (%cond_0)
        }

    Result:

        main(%a: (1, bool),
             %b: (2, 3, fp32)) {
          block0() {
            %squeeze_0: (bool) = squeeze(x=%a, name="squeeze_0")
            %cond_0: (2, 3, fp32) = identity(x=%b, name="cond_0")
          } -> (%cond_0)
        }
    """
    for f_name, f in prog.functions.items():
        num_changes = remove_vacuous_cond_block(f)
        msg = "remove_vacuous_cond: changed {} ops in function '{}'"
        logging.info(msg.format(num_changes, f_name))



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import unittest
from coremltools._deps import _HAS_ONNX, MSG_ONNX_NOT_FOUND

if _HAS_ONNX:
    import onnx
    from onnx import helper, numpy_helper, TensorProto
    from coremltools.converters.onnx._graph import Node, Graph
    from ._test_utils import (
        _onnx_create_single_node_model,
        _onnx_create_model,
        _conv_pool_output_size,
        _random_array,
    )


@unittest.skipUnless(_HAS_ONNX, MSG_ONNX_NOT_FOUND)
class NodeTest(unittest.TestCase):
    def test_create_node(self):  # type: () -> None
        model = _onnx_create_single_node_model(
            "Elu", [(1, 3, 224, 224)], [(1, 3, 224, 224)], alpha=0.5
        )
        graph = model.graph
        node = graph.node[0]
        node_ = Node.from_onnx(node)
        self.assertTrue(len(node_.inputs) == 1)
        self.assertTrue(len(node_.outputs) == 1)
        self.assertTrue(len(node_.attrs) == 1)
        self.assertTrue(node_.attrs["alpha"] == 0.5)


@unittest.skipUnless(_HAS_ONNX, MSG_ONNX_NOT_FOUND)
class GraphTest(unittest.TestCase):
    def test_create_graph(self):  # type: () -> None
        kernel_shape = (3, 2)
        strides = (2, 3)
        pads = (4, 2, 4, 2)
        dilations = (1, 2)
        group = 1
        weight = numpy_helper.from_array(_random_array((16, 3, 3, 2)), name="weight")

        input_shape = (1, 3, 224, 224)
        output_size = _conv_pool_output_size(
            input_shape, dilations, kernel_shape, pads, strides
        )

        output_shape = (1, int(weight.dims[0]), output_size[0], output_size[1])

        inputs = [("input0", input_shape)]
        outputs = [("output0", output_shape, TensorProto.FLOAT)]

        conv = helper.make_node(
            "Conv",
            inputs=[inputs[0][0], "weight"],
            outputs=["conv_output"],
            dilations=dilations,
            group=group,
            kernel_shape=kernel_shape,
            pads=pads,
            strides=strides,
        )

        relu = helper.make_node(
            "Relu", inputs=[conv.output[0]], outputs=[outputs[0][0]]
        )

        model = _onnx_create_model([conv, relu], inputs, outputs, [weight])
        graph_ = Graph.from_onnx(model.graph, onnx_ir_version=5)
        self.assertTrue(len(graph_.inputs) == 1)
        self.assertEqual(graph_.inputs[0][2], input_shape)
        self.assertTrue(len(graph_.outputs) == 1)
        self.assertEqual(graph_.outputs[0][2], output_shape)
        self.assertTrue(len(graph_.nodes) == 2)
        self.assertEqual(len(graph_.nodes[0].parents), 0)
        self.assertEqual(len(graph_.nodes[1].parents), 1)
        self.assertEqual(len(graph_.nodes[0].children), 1)
        self.assertEqual(len(graph_.nodes[1].children), 0)



# -*- coding: utf-8 -*-

#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import unittest
import pytest

pytest.importorskip("tensorflow", minversion="1.14.0")
from tensorflow.core.framework import attr_value_pb2 as attr_value
from tensorflow.core.framework import tensor_shape_pb2 as tensor_shape
from tensorflow.core.framework import types_pb2 as types

from coremltools.converters.mil.mil import types as mil_types
import coremltools.converters.mil.frontend.tensorflow.parse as parse


class TestParse(unittest.TestCase):
    def test_parse_list(self):
        def compare(expected, lst, field_name):
            attr = attr_value.AttrValue()
            field = getattr(attr.list, field_name)
            field.extend(lst)

            actual = parse.parse_attr(attr)
            self.assertEqual(expected, actual)

        compare([1, 2, 3], [1, 2, 3], "i")
        compare(["foo", "bar"], [b"foo", b"bar"], "s")

    def test_parse_scalar(self):
        def compare(expected, val, field_name):
            a = attr_value.AttrValue()
            setattr(a, field_name, val)
            actual = parse.parse_attr(a)
            self.assertEqual(expected, actual)

        compare("a String", b"a String", "s")
        compare(55, 55, "i")
        compare(True, True, "b")

        attr = attr_value.AttrValue()
        attr.f = 12.3
        self.assertAlmostEqual(12.3, parse.parse_attr(attr), places=2)

    @staticmethod
    def _attr_with_shape(dims, unknown_rank=0):
        attr = attr_value.AttrValue()
        for (dim_size, dim_name) in dims:
            tf_dim = tensor_shape.TensorShapeProto.Dim()
            tf_dim.size = dim_size
            tf_dim.name = dim_name
            attr.shape.dim.append(tf_dim)
        attr.shape.unknown_rank = unknown_rank
        return attr

    def test_parse_shape(self):
        def compare(expected, dims, unknown_rank=0):
            attr = self._attr_with_shape(dims, unknown_rank)
            actual = parse.parse_attr(attr)
            self.assertEqual(expected, actual)

        compare(None, [], 5)
        compare([100], [(100, "outer")])
        compare([1, 2, 3], [(1, "outer"), (2, "middle"), (3, "inner")])

    def test_parse_tensor(self):
        # Zero-rank tensor
        attr = attr_value.AttrValue()
        attr.tensor.version_number = 1
        attr.tensor.dtype = types.DataType.DT_INT32
        t = parse.parse_attr(attr)
        self.assertTrue(isinstance(t, mil_types.int32))
        self.assertEqual(0, t.val)

        # Non-zero rank
        attr = attr_value.AttrValue()
        attr.tensor.version_number = 1
        attr.tensor.dtype = types.DataType.DT_INT32
        shaped_attr = self._attr_with_shape([(1, "outer"), (2, "middle"), (3, "inner")])
        attr.tensor.tensor_shape.dim.extend(shaped_attr.shape.dim)
        attr.tensor.int_val.extend([55, 56, 57])

        t = parse.parse_attr(attr)
        self.assertEqual([55, 56, 57], t.val.tolist())
        self.assertEqual("tensor", mil_types.get_type_info(t).name)

        # Note that the result of t.get_primitive() is a function that returns a type
        # rather than an instance of that type as it is when the tensor has rank zero.
        self.assertTrue(isinstance(t.get_primitive()(), mil_types.int32))
        self.assertEqual((1, 2, 3), t.get_shape())

    def test_parse_type(self):
        def compare(expected, tf_type):
            attr = attr_value.AttrValue()
            attr.type = tf_type
            self.assertEqual(expected, parse.parse_attr(attr))

        compare(None, types.DataType.DT_INVALID)
        compare(mil_types.float, types.DataType.DT_FLOAT)
        compare(mil_types.double, types.DataType.DT_DOUBLE)
        compare(mil_types.int32, types.DataType.DT_INT32)
        compare(mil_types.uint8, types.DataType.DT_UINT8)
        compare(mil_types.int16, types.DataType.DT_INT16)
        compare(mil_types.int8, types.DataType.DT_INT8)
        compare(mil_types.int8, types.DataType.DT_INT8)
        compare(mil_types.str, types.DataType.DT_STRING)
        compare(None, types.DataType.DT_COMPLEX64)
        compare(mil_types.int64, types.DataType.DT_INT64)
        compare(mil_types.bool, types.DataType.DT_BOOL)
        compare(None, types.DataType.DT_QINT8)
        compare(None, types.DataType.DT_QUINT8)
        compare(None, types.DataType.DT_QINT32)
        compare(None, types.DataType.DT_BFLOAT16)
        compare(None, types.DataType.DT_QINT16)
        compare(None, types.DataType.DT_QUINT16)
        compare(mil_types.uint16, types.DataType.DT_UINT16)
        compare(None, types.DataType.DT_COMPLEX128)
        compare(mil_types.fp16, types.DataType.DT_HALF)
        compare(None, types.DataType.DT_RESOURCE)
        compare(None, types.DataType.DT_VARIANT)
        compare(mil_types.uint32, types.DataType.DT_UINT32)
        compare(mil_types.uint64, types.DataType.DT_UINT64)



import itertools

import numpy as np
import pytest
import torch

from ..internal_graph import *
from ..torchir_passes import *


def _build_flattening_test_graph():
    # This test graph is:
    #    graph(
    #        %1 : (Tensor[1, 1], (Tensor[1, 2], Tensor[1, 3]))
    #    ):
    #    %2, %3 = tupleunpack[](%1)
    #    %4, %5 = tupleunpack[](%3)
    #    %6 = tupleconstruct[](%2, %4)
    #    %7 = tupleconstruct[](%6, %5)
    #    return (%7)
    #
    # And if you were to run the graph it would turn
    #    (a, (b, c))
    # into
    #    ((a, b), c)

    graph_params = {}
    graph_inputs = OrderedDict()
    graph_inputs["1"] = (
        torch.rand(1, 1),
        (
            torch.rand(1, 2),
            torch.rand(1, 3),
        ),
    )
    graph_nodes = [
        InternalTorchIRNode(
            inputs=["1"],
            outputs=["2", "3"],
            kind="tupleunpack",
        ),
        InternalTorchIRNode(
            inputs=["3"],
            outputs=["4", "5"],
            kind="tupleunpack",
        ),
        InternalTorchIRNode(
            inputs=["2", "4"],
            outputs=["6"],
            kind="tupleconstruct",
        ),
        InternalTorchIRNode(
            inputs=["6", "5"],
            outputs=["7"],
            kind="tupleconstruct",
        ),
    ]
    graph_outputs = ["7"]

    return InternalTorchIRGraph(
        nodes=graph_nodes,
        params=graph_params,
        inputs=graph_inputs,
        outputs=graph_outputs,
    )


class TestTorchPasses:
    """Class containing tests for InternalTorchIR optimization passes.
    """

    @pytest.fixture
    def set_random_seeds(self):
        torch.manual_seed(1)
        np.random.seed(1)

    def test_flatten_input_values(self):
        graph = _build_flattening_test_graph()

        flatten_graph_input_values(graph)

        # The graph input tuple should have been flattened.
        np.testing.assert_equal(len(graph.inputs.keys()), 3)
        # Tuple flattening should introduce two new ops.
        np.testing.assert_equal(len(graph.nodes), 6)
        # The new ops at the beginning of the graph should be a tupleconstruct.
        np.testing.assert_equal(graph.nodes[0].kind, "tupleconstruct")
        np.testing.assert_equal(graph.nodes[1].kind, "tupleconstruct")
        # The inputs to the tupleconstructs should be the new flattened inputs.
        input_names = [k for k in graph.inputs.keys()]
        np.testing.assert_equal(input_names[1:], graph.nodes[0].inputs)
        np.testing.assert_equal(input_names[0], graph.nodes[1].inputs[0])
        np.testing.assert_equal(graph.nodes[0].outputs[0], graph.nodes[1].inputs[1])
        # The last inserted tuple construct should produce the input for the
        # next op.
        np.testing.assert_equal(graph.nodes[1].outputs[0], graph.nodes[2].inputs[0])

    def test_flatten_output_values(self):
        graph = _build_flattening_test_graph()

        flatten_graph_output_values(graph)

        # The graph output tuple should have been flattened.
        np.testing.assert_equal(len(graph.outputs), 3)
        # The outputs of the graph should come from intermediate ops.
        np.testing.assert_equal(graph.outputs[0], graph.nodes[0].outputs[0])
        np.testing.assert_equal(graph.outputs[1], graph.nodes[1].outputs[0])
        np.testing.assert_equal(graph.outputs[2], graph.nodes[1].outputs[1])

    def test_transform_inplace_ops_graph(self):
        # The test graph is:
        #    graph(
        #        %x : Tensor[1],
        #    ):
        #      %1 = constant[value=0]()
        #      %2 = constant[value=10]()
        #      %3 = listconstruct[](%1)
        #      %4 = append[](%3, %2)
        #      return (%3)
        graph_params = {}
        graph_inputs = OrderedDict()
        graph_inputs["x"] = torch.rand(1)
        graph_nodes = [
            InternalTorchIRNode(
                inputs=[],
                attr={"value": 0},
                outputs=["1"],
                kind="constant",
            ),
            InternalTorchIRNode(
                inputs=[],
                attr={"value": 10},
                outputs=["2"],
                kind="constant",
            ),
            InternalTorchIRNode(
                inputs=["1"],
                outputs=["3"],
                kind="listconstruct",
            ),
            InternalTorchIRNode(
                inputs=["3", "2"],
                outputs=["4"],
                kind="append",
            ),
        ]
        graph_outputs = ["3"]
        graph = InternalTorchIRGraph(
            nodes=graph_nodes,
            params=graph_params,
            inputs=graph_inputs,
            outputs=graph_outputs,
        )
        for node in graph.nodes:
            node.parent = graph

        transform_inplace_ops(graph)

        np.testing.assert_equal(len(graph.outputs), 1)
        np.testing.assert_equal(graph.outputs[0], graph.nodes[-1].outputs[0])


    def test_transform_inplace_ops_loop(self):
        # The test graph is:
        #    graph(
        #        %x : Tensor[1],
        #    ):
        #      %1 = constant[value=True]()
        #      %2 = constant[value=-1]()
        #      %3 = constant[value=10]()
        #      %4 = listconstruct[](%2)
        #       = loop[](%3, %1)
        #        block(%i.1):
        #          %6 = append[](%4, %i.1)
        #        return (%1)
        #    return (%4)
        graph_params = {}
        graph_inputs = OrderedDict()
        graph_inputs["x"] = torch.rand(1)
        loop_block = InternalTorchIRBlock(
            inputs=["i.1"],
            outputs=["1"],
            nodes=[
                InternalTorchIRNode(
                    inputs=["4", "i.1"],
                    outputs=["6"],
                    kind="append",
                ),
            ],
        )
        loop_block.nodes[0].parent = loop_block
        loop_node = InternalTorchIRNode(
            inputs=["3", "1"],
            outputs=[],
            kind="loop",
            blocks=[loop_block],
        )
        loop_block.parent = loop_node
        graph_nodes = [
            InternalTorchIRNode(
                inputs=[],
                attr={"value": True},
                outputs=["1"],
                kind="constant",
            ),
            InternalTorchIRNode(
                inputs=[],
                attr={"value": -1},
                outputs=["2"],
                kind="constant",
            ),
            InternalTorchIRNode(
                inputs=[],
                attr={"value": 10},
                outputs=["3"],
                kind="constant",
            ),
            InternalTorchIRNode(
                inputs=["2"],
                outputs=["4"],
                kind="listconstruct",
            ),
            loop_node,
        ]
        graph_outputs = ["4"]
        graph = InternalTorchIRGraph(
            nodes=graph_nodes,
            params=graph_params,
            inputs=graph_inputs,
            outputs=graph_outputs,
        )
        for node in graph.nodes:
            node.parent = graph

        transform_inplace_ops(graph)

        # There should be an additional input to the loop.
        np.testing.assert_equal(len(loop_node.inputs), 3)
        # That input should be the output of the previous op.
        np.testing.assert_equal(loop_node.inputs[2], graph.nodes[3].outputs[0])
        # The loop block should have an additional input.
        np.testing.assert_equal(len(loop_block.inputs), 2)
        # The loop block's new input should be the input to append.
        np.testing.assert_equal(loop_block.inputs[1], loop_block.nodes[0].inputs[0])
        # The loop block should have an additional output.
        np.testing.assert_equal(len(loop_block.outputs), 2)
        # Append's output should be returned from the loop block.
        np.testing.assert_equal(loop_block.outputs[1], loop_block.nodes[0].outputs[0])
        # The loop should now have an output.
        np.testing.assert_equal(len(loop_node.outputs), 1)
        # The loop's name should now be the name of its output.
        np.testing.assert_equal(loop_node.name, loop_node.outputs[0])
        # That graph output should now be the output of the graph.
        np.testing.assert_equal(loop_node.outputs[0], graph.outputs[0])


    @pytest.mark.xfail(reason="rdar://64235006")
    def test_transform_inplace_ops_if(self):
        # The test graph is:
        #    graph(
        #        %x : Tensor[1],
        #    ):
        #      %1 = constant[value=True]()
        #      %2 = constant[value=0]()
        #      %3 = constant[value=1]()
        #      %4 = listconstruct[](%2)
        #       = if[](%1)
        #        block0():
        #          %5 = append[](%4, %3)
        #        return ()
        #        block1():
        #          %6 = append[](%4, %2)
        #        return ()
        #    return (%4)
        graph_params = {}
        graph_inputs = OrderedDict()
        graph_inputs["x"] = torch.rand(1)
        if_true_block = InternalTorchIRBlock(
            inputs=[],
            outputs=[],
            nodes=[
                InternalTorchIRNode(
                    inputs=["4", "3"],
                    outputs=["5"],
                    kind="append",
                ),
            ],
        )
        if_true_block.nodes[0].parent = if_true_block
        if_false_block = InternalTorchIRBlock(
            inputs=[],
            outputs=[],
            nodes=[
                InternalTorchIRNode(
                    inputs=["4", "2"],
                    outputs=["6"],
                    kind="append",
                ),
            ],
        )
        if_false_block.nodes[0].parent = if_false_block
        if_node = InternalTorchIRNode(
            inputs=["1"],
            outputs=[],
            kind="if",
            blocks=[if_true_block, if_false_block],
        )
        if_true_block.parent = if_node
        if_false_block.parent = if_node
        graph_nodes = [
            InternalTorchIRNode(
                inputs=[],
                attr={"value": True},
                outputs=["1"],
                kind="constant",
            ),
            InternalTorchIRNode(
                inputs=[],
                attr={"value": 0},
                outputs=["2"],
                kind="constant",
            ),
            InternalTorchIRNode(
                inputs=[],
                attr={"value": 1},
                outputs=["3"],
                kind="constant",
            ),
            InternalTorchIRNode(
                inputs=["2"],
                outputs=["4"],
                kind="listconstruct",
            ),
            if_node,
        ]
        graph_outputs = ["4"]
        graph = InternalTorchIRGraph(
            nodes=graph_nodes,
            params=graph_params,
            inputs=graph_inputs,
            outputs=graph_outputs,
        )
        for node in graph.nodes:
            node.parent = graph

        transform_inplace_ops(graph)

        # The true block should now have an output.
        np.testing.assert_equal(len(if_true_block.outputs), 1)
        # The true block should output the result of the append op.
        np.testing.assert_equal(if_true_block.outputs[0], if_true_block.nodes[0].outputs[0])
        # The false block should now have an output.
        np.testing.assert_equal(len(if_false_block.outputs), 1)
        # The false block should output the result of the append op.
        np.testing.assert_equal(if_false_block.outputs[0], if_false_block.nodes[0].outputs[0])
        # The if op should have an additional output.
        np.testing.assert_equal(len(if_node.outputs), 1)
        # The if's name should now be the name of its output.
        np.testing.assert_equal(if_node.name, if_node.outputs[0])
        # The graph output should be the if op output.
        np.testing.assert_equal(if_node.outputs[0], graph.outputs[0])